// Copyright (c) 2004, 2018, Oracle and/or its affiliates. All rights reserved.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License, version 2.0, as
// published by the Free Software Foundation.
//
// This program is also distributed with certain software (including
// but not limited to OpenSSL) that is licensed under separate terms,
// as designated in a particular file or component or in included license
// documentation.  The authors of MySQL hereby grant you an
// additional permission to link the program and your derivative works
// with the separately licensed software that they have included with
// MySQL.
//
// Without limiting anything contained in the foregoing, this file,
// which is part of MySQL Connector/NET, is also subject to the
// Universal FOSS Exception, version 1.0, a copy of which can be found at
// http://oss.oracle.com/licenses/universal-foss-exception.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
// See the GNU General Public License, version 2.0, for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software Foundation, Inc.,
// 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA


using MySql.Data.MySqlClient;
using Org.BouncyCastle.Crypto;
using Org.BouncyCastle.OpenSsl;
using Org.BouncyCastle.X509;
using Org.BouncyCastle.Security.Certificates;
using Org.BouncyCastle.Security;
using System;
using System.Collections.Generic;
using System.IO;
using System.Net.Security;
using System.Security.Authentication;
using System.Security.Cryptography.X509Certificates;
using System.Text;

namespace MySql.Data.Common
{
  /// <summary>
  /// Handles SSL connections for the Classic and X protocols.
  /// </summary>
  internal class Ssl
  {
    #region Fields

    /// <summary>
    /// Contains the connection options provided by the user.
    /// </summary>
    private MySqlConnectionStringBuilder _settings;

    /// <summary>
    /// A flag to establish how certificates are to be treated and validated.
    /// </summary>
    private bool _treatCertificatesAsPemFormat;

    /// <summary>
    /// Defines the supported TLS protocols.
    /// </summary>
    private static SslProtocols[] tlsProtocols = new SslProtocols[] { SslProtocols.Tls12, SslProtocols.Tls11 };
    private static Dictionary<string, SslProtocols> tlsConnectionRef = new Dictionary<string, SslProtocols>();
    private static Dictionary<string, int> tlsRetry = new Dictionary<string, int>();
    private static Object thisLock = new Object();

    #endregion

    public Ssl(MySqlConnectionStringBuilder settings)
    {
      this._settings = settings;
      // Set default value to true since PEM files is the standard for MySQL SSL certificates.
      _treatCertificatesAsPemFormat = true;
    }

    public Ssl(string server, MySqlSslMode sslMode, string certificateFile, MySqlCertificateStoreLocation certificateStoreLocation,
        string certificatePassword, string certificateThumbprint, string sslCa, string sslCert, string sslKey)
    {
      this._settings = new MySqlConnectionStringBuilder()
      {
        Server = server,
        SslMode = sslMode,
        CertificateFile = certificateFile,
        CertificateStoreLocation = certificateStoreLocation,
        CertificatePassword = certificatePassword,
        CertificateThumbprint = certificateThumbprint,
        SslCa = sslCa,
        SslCert = sslCert,
        SslKey = sslKey
      };
      // Set default value to true since PEM files is the standard for MySQL SSL certificates.
      _treatCertificatesAsPemFormat = true;
    }

    /// <summary>
    /// Retrieves a collection containing the client SSL PFX certificates.
    /// </summary>
    /// <remarks>Dependent on connection string settings.
    /// Either file or store based certificates are used.</remarks>
    private X509CertificateCollection GetPFXClientCertificates()
    {
      X509CertificateCollection certs = new X509CertificateCollection();

      // Check for file-based certificate
      if (_settings.CertificateFile != null)
      {
        X509Certificate2 clientCert = new X509Certificate2(_settings.CertificateFile,
            _settings.CertificatePassword);
        certs.Add(clientCert);
        return certs;
      }

      if (_settings.CertificateStoreLocation == MySqlCertificateStoreLocation.None)
        return certs;

      StoreLocation location =
          (_settings.CertificateStoreLocation == MySqlCertificateStoreLocation.CurrentUser) ?
          StoreLocation.CurrentUser : StoreLocation.LocalMachine;

      // Check for store-based certificate
      X509Store store = new X509Store(StoreName.My, location);
      store.Open(OpenFlags.ReadOnly | OpenFlags.OpenExistingOnly);


      if (_settings.CertificateThumbprint == null)
      {
        // Return all certificates from the store.
        certs.AddRange(store.Certificates);
        return certs;
      }

      // Find certificate with given thumbprint
      certs.AddRange(store.Certificates.Find(X509FindType.FindByThumbprint,
                _settings.CertificateThumbprint, true));

      if (certs.Count == 0)
      {
        throw new MySqlException("Certificate with Thumbprint " +
           _settings.CertificateThumbprint + " not found");
      }
      return certs;
    }

    /// <summary>
    /// Initiates the SSL connection.
    /// </summary>
    /// <param name="baseStream">The base stream.</param>
    /// <param name="encoding">The encoding used in the SSL connection.</param>
    /// <param name="connectionString">The connection string used to establish the connection.</param>
    /// <returns>A <see cref="MySqlStream"/> instance ready to initiate an SSL connection.</returns>
    public MySqlStream StartSSL(ref Stream baseStream, Encoding encoding, string connectionString)
    {
      // If SslCa connection option was provided, check for the file extension as it can also be set as a PFX file.
      if (_settings.SslCa != null)
      {
        var fileExtension = GetCertificateFileExtension(_settings.SslCa, true);

        if (fileExtension != null)
          _treatCertificatesAsPemFormat = fileExtension != "pfx";
      }

      RemoteCertificateValidationCallback sslValidateCallback =
          new RemoteCertificateValidationCallback(ServerCheckValidation);
      SslStream ss = new SslStream(baseStream, false, sslValidateCallback, null);
      X509CertificateCollection certs = _treatCertificatesAsPemFormat
        ? new X509CertificateCollection()
        : GetPFXClientCertificates();

      string connectionId = connectionString.GetHashCode().ToString();
      SslProtocols tlsProtocol = SslProtocols.Tls;

      lock (thisLock)
      {
        if (tlsConnectionRef.ContainsKey(connectionId))
        {
          tlsProtocol = tlsConnectionRef[connectionId];
        }
        else
        {
          if (!tlsRetry.ContainsKey(connectionId))
          {
            tlsRetry[connectionId] = 0;
          }
          for (int i = tlsRetry[connectionId]; i < tlsProtocols.Length; i++)
          {
            tlsProtocol |= tlsProtocols[i];
          }
        }
        try
        {
          ss.AuthenticateAsClientAsync(_settings.Server, certs, tlsProtocol, false).Wait();
          tlsConnectionRef[connectionId] = tlsProtocol;
          tlsRetry.Remove(connectionId);
        }
        catch (AggregateException ex)
        {
          if (ex.GetBaseException() is IOException)
          {
            tlsConnectionRef.Remove(connectionId);
            if (tlsRetry.ContainsKey(connectionId))
            {
              if (tlsRetry[connectionId] > tlsProtocols.Length)
                throw new MySqlException(Resources.SslConnectionError, ex);
              tlsRetry[connectionId] += 1;
            }
          }
          throw ex.GetBaseException();
        }
      }

      baseStream = ss;
      MySqlStream stream = new MySqlStream(ss, encoding, false);
      stream.SequenceByte = 2;

      return stream;
    }

    /// <summary>
    /// Verifies the SSL certificates used for authentication.
    /// </summary>
    /// <param name="sender">An object that contains state information for this validation.</param>
    /// <param name="certificate">The MySQL server certificate used to authenticate the remote party.</param>
    /// <param name="chain">The chain of certificate authorities associated with the remote certificate.</param>
    /// <param name="sslPolicyErrors">One or more errors associated with the remote certificate.</param>
    /// <returns><c>true</c> if no errors were found based on the selected SSL mode; <c>false</c>, otherwise.</returns>
    private bool ServerCheckValidation(object sender, System.Security.Cryptography.X509Certificates.X509Certificate certificate,
                                              X509Chain chain, SslPolicyErrors sslPolicyErrors)
    {
      if (sslPolicyErrors == SslPolicyErrors.None)
        return true;

      if (_settings.SslMode == MySqlSslMode.Required ||
          _settings.SslMode == MySqlSslMode.Preferred)
      {
        // Tolerate all certificate errors.
        return true;
      }

      // Validate PEM certificates using Bouncy Castle.
      if (_treatCertificatesAsPemFormat)
      {
        if (_settings.SslMode >= MySqlSslMode.VerifyCA)
        {
          VerifyEmptyOrWhitespaceSslConnectionOption(_settings.SslCa, nameof(_settings.SslCa));
          var sslCA = ReadSslCertificate(_settings.SslCa);
          VerifyIssuer(sslCA, certificate);
          VerifyDates(sslCA);
          VerifyCAStatus(sslCA, true);
#if NET452
          VerifySignature(sslCA, DotNetUtilities.FromX509Certificate(certificate));
#elif !NETSTANDARD1_6
          VerifySignature(sslCA, new X509CertificateParser().ReadCertificate(certificate.GetRawCertData()));
#endif
        }

        if (_settings.SslMode == MySqlSslMode.VerifyFull)
        {
          VerifyEmptyOrWhitespaceSslConnectionOption(_settings.SslCert, nameof(_settings.SslCert));
          var sslCert = ReadSslCertificate(_settings.SslCert);
          VerifyDates(sslCert);
          VerifyCAStatus(sslCert, false);

          VerifyEmptyOrWhitespaceSslConnectionOption(_settings.SslKey, nameof(_settings.SslKey));
          var sslKey = ReadKey(_settings.SslKey);
          VerifyKeyCorrespondsToCertificateKey(sslCert, sslKey);
        }

        return true;
      }
      // Validate PFX certificate errors.
      else if (_settings.SslMode == MySqlSslMode.VerifyCA &&
          sslPolicyErrors == SslPolicyErrors.RemoteCertificateNameMismatch)
      {
        // Tolerate name mismatch in certificate, if full validation is not requested.
        return true;
      }

      return false;
    }

    #region Certificate Readers

    /// <summary>
    /// Reads the specified file as a byte array.
    /// </summary>
    /// <param name="fileName">The path of the file to read.</param>
    /// <returns>A byte array representing the read file.</returns>
    private static byte[] GetBuffer(string filePath)
    {
      byte[] buffer;
      if (filePath == null)
      {
        throw new ArgumentNullException(nameof(filePath));
      }

      try
      {
        FileStream stream = File.Open(filePath, FileMode.Open, FileAccess.Read, FileShare.ReadWrite);
        buffer = new byte[stream.Length];
        int offset = stream.Read(buffer, 0, buffer.Length);
        while (true)
        {
          if (offset >= stream.Length)
          {
#if NETSTANDARD1_6
            stream.Dispose();
#else
            stream.Close();
#endif
            break;
          }
          offset += stream.Read(buffer, offset, buffer.Length - offset);
        }
      }
      catch (Exception)
      {
        throw new MySqlException(Resources.SslConnectionError, new FileNotFoundException(Resources.FileNotFound, filePath));
      }

      return buffer;
    }

    /// <summary>
    /// Gets the extension of the specified file.
    /// </summary>
    /// <param name="filePath">The path of the file.</param>
    /// <param name="toLowerCase">Flag to indicate if the result should be converted to lower case.</param>
    /// <remarks>The . character is ommited from the result.</remarks>
    /// <returns></returns>
    private string GetCertificateFileExtension(string filePath, bool toLowerCase)
    {
      if (filePath == null || !File.Exists(filePath))
        return null;

      var extension = Path.GetExtension(filePath);
      extension = string.IsNullOrEmpty(extension)
        ? null
        : extension.Substring(extension.IndexOf(".") + 1);

      return toLowerCase
        ? extension.ToLowerInvariant()
        : extension;
    }

    /// <summary>
    /// Reads the SSL certificate file.
    /// </summary>
    /// <param name="filePath">The path to the certificate file.</param>
    /// <returns>A <see cref="Org.BouncyCastle.X509.X509Certificate"/> instance representing the SSL certificate file.</returns>
    private Org.BouncyCastle.X509.X509Certificate ReadSslCertificate(string filePath)
    {
      byte[] buffer = GetBuffer(filePath);
      var PR = new PemReader(new StreamReader(new MemoryStream(buffer)));

      try
      {
        var certificate = (Org.BouncyCastle.X509.X509Certificate)PR.ReadObject();
        if (certificate == null)
          throw new InvalidCastException();

        return certificate;
      }
      catch (InvalidCastException)
      {
        throw new MySqlException(Resources.SslConnectionError, new Exception(Resources.FileIsNotACertificate));
      }
    }

    /// <summary>
    /// Reads the SSL certificate key file.
    /// </summary>
    /// <param name="filePath">The path to the certificate key file.</param>
    /// <returns>A <see cref="AsymmetricCipherKeyPair"/> instance representing the SSL certificate key file.</returns>
    private AsymmetricCipherKeyPair ReadKey(string filePath)
    {
      byte[] buffer = GetBuffer(filePath);
      var PR = new PemReader(new StreamReader(new MemoryStream(buffer)));

      try
      {
        var key = (AsymmetricCipherKeyPair)PR.ReadObject();
        if (key == null)
          throw new InvalidCastException();

        return key;
      }
      catch (InvalidCastException)
      {
        throw new MySqlException(Resources.SslConnectionError, new Exception(Resources.FileIsNotAKey));
      }
    }

    #endregion

    #region Certificate Veryfiers

    /// <summary>
    /// Verifies that the certificate has not yet expired.
    /// </summary>
    /// <param name="certificate">The certificate to verify.</param>
    private void VerifyDates(Org.BouncyCastle.X509.X509Certificate certificate)
    {
      try
      {
        certificate.CheckValidity();
      }
      catch (CertificateExpiredException exception)
      {
        throw new MySqlException(Resources.SslConnectionError, exception);
      }
      catch (CertificateNotYetValidException exception)
      {
        throw new MySqlException(Resources.SslConnectionError, exception);
      }
    }

    /// <summary>
    /// Verifies a certificate CA status.
    /// </summary>
    /// <param name="certificate">The certificate to validate.</param>
    /// <param name="expectedCAStatus">A flag indicating the expected CA status.</param>
    private void VerifyCAStatus(Org.BouncyCastle.X509.X509Certificate certificate, bool expectedCAStatus)
    {
      var certificatePathLength = -1;
      bool? isCA = IsCA(certificate, out certificatePathLength);
      if (isCA == true && !expectedCAStatus)
        throw new MySqlException(Resources.SslConnectionError, new Exception(Resources.InvalidSslCertificate));
      else if ((isCA == false || isCA == null) && expectedCAStatus)
        throw new MySqlException(Resources.SslConnectionError, new Exception(Resources.SslCertificateIsNotCA));
    }

    /// <summary>
    /// Verifies that the certificate was signed using the private key that corresponds to the specified public key
    /// </summary>
    /// <param name="certificate">The client side certificate containing the public key.</param>
    /// <param name="serverCertificate">The server certificate.</param>
    private void VerifySignature(Org.BouncyCastle.X509.X509Certificate certificate, Org.BouncyCastle.X509.X509Certificate serverCertificate)
    {
      VerifySignature(serverCertificate, certificate.GetPublicKey());
    }

    private void VerifySignatureUsingKey(Org.BouncyCastle.X509.X509Certificate certificate, AsymmetricCipherKeyPair key)
    {
      VerifySignature(certificate, key.Public);
    }

    private void VerifySignature(Org.BouncyCastle.X509.X509Certificate certificate, AsymmetricKeyParameter key)
    {
      try
      {
        certificate.Verify(key);
      }
      catch (InvalidKeyException exception)
      {
        throw new Exception(Resources.InvalidCertificateKey, exception);
      }
      catch (SignatureException exception)
      {
        throw new Exception(Resources.InvalidSslCertificateSignature, exception);
      }
      catch (CertificateException exception)
      {
        throw new Exception(Resources.EncodingError, exception);
      }
      catch (Exception exception)
      {
        throw new Exception(Resources.InvalidSslCertificateSignatureGeneral, exception);
      }
    }

    /// <summary>
    /// Verifies that no SSL policy errors regarding the identitfy of the host were raised.
    /// </summary>
    /// <param name="sslPolicyErrors">A <see cref="SslPolicyErrors"/> instance set with the raised SSL errors.</param>
    private void VerifyIdentity(SslPolicyErrors sslPolicyErrors)
    {
      if (sslPolicyErrors == SslPolicyErrors.RemoteCertificateNameMismatch)
        throw new MySqlException(Resources.SslConnectionError, new Exception(Resources.SslCertificateHostNameMismatch));
    }

    /// <summary>
    /// Verifies that the CA by comparing the CA certificate issuer and the server certificate issuer.
    /// </summary>
    /// <param name="CACertificate">The CA certificate.</param>
    /// <param name="serverCertificate">The server certificate.</param>
    private void VerifyIssuer(Org.BouncyCastle.X509.X509Certificate CACertificate, System.Security.Cryptography.X509Certificates.X509Certificate serverCertificate)
    {
      if (CACertificate.IssuerDN.ToString() != serverCertificate.Issuer)
        throw new MySqlException(Resources.SslConnectionError, new Exception(Resources.SslCertificateCAMismatch));
    }

    private void VerifyKeyCorrespondsToCertificateKey(Org.BouncyCastle.X509.X509Certificate certificate, AsymmetricCipherKeyPair key)
    {
      var certificateKey = certificate.GetPublicKey().ToString();
      if (!string.IsNullOrEmpty(certificateKey) && certificateKey != key.Public.ToString())
        throw new InvalidKeyException();
    }

    /// Validates that the certificate provided is a CA certificate.
    /// </summary>
    /// <param name="certificate">The certificate to validate.</param>
    /// <param name="certificationPathLength">The allowed certification path length.</param>
    /// <returns><c>null</c> if the certificate info does not allow to determine the CA status;
    /// otherwise, a boolean value indicating the CA status.</null></returns>
    private bool? IsCA(Org.BouncyCastle.X509.X509Certificate certificate, out int certificationPathLength)
    {
      // If certificate version equal to 3 then the isCA property can be retrieved.
      if (certificate.Version == 3)
      {
        // A value of -1 indicates certificate is not a CA.
        // A value of Integer.MAX_VALUE indicates there is no limit on the allowed length of the certification path.
        certificationPathLength = certificate.GetBasicConstraints();
        return certificationPathLength != -1;
      }

      certificationPathLength = -1;
      return null;
    }

    #endregion

    /// <summary>
    /// Raises an exception if the specified connection option is null, empty or whitespace.
    /// </summary>
    /// <param name="connectionOption">The connection option to verify.</param>
    private void VerifyEmptyOrWhitespaceSslConnectionOption(string connectionOption, string connectionOptionName)
    {
      if (string.IsNullOrWhiteSpace(connectionOption))
        throw new MySqlException(
          Resources.SslConnectionError,
          new FileNotFoundException(string.Format(Resources.FilePathNotSet, connectionOptionName)));
    }
  }
}
